use std::{fs}; //,io};
use std::fs::OpenOptions;
use std::path::Path;
use std::io::{Write};
use std::str;
use anyhow::{anyhow,Result};
// use tokio::runtime::Handle;
use std::process::{Command};

use app_env::{
	appenv::AppEnv,
  // config::{Config}
};
use crate::clouds::defs::{
  Cloud,
};

use crate::defs::ProviderName;

use crate::clouds::monitor_rules::{
	MonitorRules,
	MonitorRule,
	RuleContext,
	RuleOperator,
	MonitorAction,
};

pub fn write_str_data(output_path: String, data_str: String, msg: String) -> Result<()> {
	if Path::new(&output_path).exists() {
		fs::remove_file(&output_path)?;
	}
	let mut file = OpenOptions::new().write(true).create(true).open(&output_path)?;
	file.write_all(data_str.as_bytes())?;
	if ! msg.is_empty() {
		let now = chrono::Utc::now().timestamp();
		let out = format!("{}: [{}] -> {}\n",&now,&msg,&output_path);
		println!("{}",&out);
	}
	Ok(())
}
pub async fn on_action_server(task: &str, srvr: &str, provider: &str, args: &str, rule: &MonitorRule) -> Result<()> {
	let provider_name = ProviderName::set_provider(provider.to_owned());
	match provider_name {
	  ProviderName::upcloud => {
			let	cmd = format!("upclapi -c {}server -id {} {}",&task,&srvr,&args);
			println!("action {}: {} on {} {}",&task,&srvr,&provider_name,&args);
			let res = run_command(&cmd).unwrap_or_else(|e|{
				eprintln!("Error {}: {}",&cmd,e);	
				String::from("")
			});
			println!("{}",&res);
			Ok(())
		},
		 _ => Err(anyhow!("Provider {} not defined", &provider_name))
	}
}
pub async fn on_action_notify(chnl: &str, msg: &str, args: &str, rule: &MonitorRule) -> Result<()> {
	Ok(())
}
fn get_server(target: &str, val: &str) -> String {
	let srvr: String;
	if val == "$target" || val == "$server" {
		srvr = target.to_owned();
	} else {
		srvr = val.to_owned();
	}
	srvr
}
pub async fn on_monitor_action(action: &MonitorAction, target: &str, rule: &MonitorRule) -> Result<()> {
  match action {
		MonitorAction::None => return Ok(()),
		MonitorAction::Start(v,p,a) => 
			on_action_server("start",&get_server(target, v),&p,&a,&rule).await?,
		MonitorAction::Stop(v,p,a) => 
			on_action_server("stop",&get_server(target, v),&p,&a,&rule).await?,
		MonitorAction::Restart(v,p,a) => 
			on_action_server("restart",&get_server(target, v),&p,&a,&rule).await?,
		MonitorAction::Notify(v,p,a) => on_action_notify(&v,&p,&a,&rule).await?,
	};
	Ok(())
}
pub fn on_monitor_operator(operator: &RuleOperator, target: &str, value: &str) -> bool {
  match operator {
		RuleOperator::None => true,
		RuleOperator::Equal => target == value,
		RuleOperator::NotEqual=> target != value,
		RuleOperator::Contains => target.contains(value),
		RuleOperator::NotContains => !target.contains(value),
	}
}
pub async fn on_server_rule(srvr: &str, out_cmd: &str, rule: &MonitorRule) -> Result<()> {
	Ok(())
}
pub async fn on_service_rule(srvc: &str, out_cmd: &str, rule: &MonitorRule) -> Result<()> {
	println!("Service: {} -> {}",&srvc,&out_cmd);
	if !out_cmd.is_empty() {
		let mut targets = out_cmd.split_whitespace();
		while let Some(target) = targets.next() {
			// println!("target: {}",&target);
			for action in &rule.actions {
				on_monitor_action(action, &target, &rule).await?;
			}
		};
	}
	Ok(())
}
pub async fn on_ip_rule(ip: &str, out_cmd: &str, rule: &MonitorRule) -> Result<()> {

	// tokio::spawn(async move {run_on_rule(rule.to_owned()).await; });
	Ok(())
}
pub fn run_command(str_cmd: &str) -> Result<String> {
  let mut args: Vec<&str> = str_cmd.split(" ").collect();
	let cmd = args[0].to_owned();
	args.remove(0);
	// println!("cmd: {}, args: {:?}",&cmd,&args);
	let result = Command::new(&cmd)
	.args(&args)
	.output()?;
 	// dbg!(&result);
	match &result.status.code() {
		Some(code) =>
			if format!("{}",code) == "0" {
				let mut res=String::from(str::from_utf8(&result.stdout).unwrap_or_else(|_| ""));
				if res.ends_with('\n') {
					res.pop();
				}
				res = res.replace('\n'," ");
				Ok(res)
		  } else {
				let err = str::from_utf8(&result.stderr).unwrap_or_else(|_| "");
		    Err(anyhow!("Running {}: failed:\n {}",&cmd,&err))
			}
		None => Ok(String::from(""))
	}
}
pub fn run_on_selector(rule: &MonitorRule) -> Result<String> {
	let ctx_val = rule.context.get_value();
	let target: Vec<&str> = ctx_val.split("/").collect();
	let srv_path = match target.len() {
		2 => String::from(target[1]),
		_ => String::from(""),
	};
	let mut cmd = match &rule.context {
		RuleContext::Server(v) => String::from(v),
		RuleContext::Service(v) =>
			if v.contains("kubernetes") {
				String::from("kubectl")
			} else {
				String::from(v)
			}
		,
		RuleContext::Ip(v) => String::from(v),
		RuleContext::None => String::from(""),
	};
	if let Some(ctxpath) = rule.selector.get("ctxpath") {
		cmd = format!("{}{} {}",&ctxpath,&srv_path,&cmd);
	}
	if let Some(run) = rule.selector.get("run") {
		cmd = format!("{} {}",&run,&cmd);
	}
	if let Some(typ) = rule.selector.get("type") {
		cmd = format!("{} get {}",&cmd,&typ);
	}
	if let Some(ns) = rule.selector.get("ns") {
		cmd = format!("{} -n {}",&cmd,&ns);
	}
	if let Some(custom) = rule.selector.get("custom") {
		cmd = format!("{} {}",&cmd,&custom);
	}
	if let Some(pat) = rule.selector.get("match") {
		cmd = format!("{} | grep {}",&cmd,&pat);
	}
	if let Some(awk) = rule.selector.get("awk") {
		cmd = format!("{} | awk '{}'",&cmd,&awk);
	}
	if let Some(sort) = rule.selector.get("sort") {
		cmd = format!("{} | sort {}",&cmd,&sort);
	}
	if cmd.len() > 0 {
		run_command(&cmd)
	}	else {
		Err(anyhow!("Selector: failed:\n {}",&rule.id))
	}
}
pub async fn run_monitor(monitor_rules: MonitorRules, _cloud: Cloud, _app_env: AppEnv) -> Result<()> {
	// dbg!(&monitor_rules); 
	println!("Run {}: {}",&monitor_rules.name,&monitor_rules.description);
	for (idx,rule) in monitor_rules.rules.iter().enumerate() {
		match rule.schedule.as_str() {
			"inmediate" => {
	      // dbg!(&rule);
	 			println!("{} [{}] {}: {}",&idx,&rule.id,&rule.name,&rule.description);
				let out_cmd: String;
				if rule.command.is_empty() {
          out_cmd=run_on_selector(&rule).unwrap_or_else(|e| { 
						eprintln!("Error on selector {}: {}",&rule.id,e);
						String::from("?")
					});
				} else {
          out_cmd=run_command(&rule.command).unwrap_or_else(|e| { 
						eprintln!("Error on command {}: {}",&rule.command,e);
						String::from("?")
					});
				}
				if out_cmd.as_str() == "?" {
					continue;
				}
				if ! on_monitor_operator(&rule.operator,&out_cmd, &rule.value) {
					continue;
				}
				match &rule.context {
					RuleContext::Server(v) => {
						on_server_rule(&v, &out_cmd, &rule).await.unwrap_or_else(|e|{
							eprintln!("Error rule {} on {}: {}",&rule.id,&v,e);	
						});	
					},
					RuleContext::None => { continue
					},
					RuleContext::Service(v) => {
						on_service_rule(&v, &out_cmd, &rule).await.unwrap_or_else(|e|{
							eprintln!("Error rule {} on {}: {}",&rule.id,&v,e);	
						});	
					},
					RuleContext::Ip(v) => {
						on_ip_rule(&v, &out_cmd, &rule).await.unwrap_or_else(|e|{
							eprintln!("Error rule {} on {}: {}",&rule.id,&v,e);	
						});	
					},
				}
			},
			_ => continue, 
		};
	}
	
  Ok(())
}