use serde::{Serialize,Deserialize};
use std::collections::HashMap;
use std::fs; //, io};
use std::path::Path;
use anyhow::{Result,Context};

use crate::clouds::defs::{Cloud,TskSrvc};

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct Haproxy {
  pub mode: String,
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct Coredns {
  pub port: u16,
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct Etcd {
  pub nodePort: u16,
  pub port: u16,
  pub name: String,
  pub cn: String,
  pub sslMode: String,
  pub dataDir: String,
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct Keepalived {
  pub mainPort: u16,
  pub mainIP: String,
  pub services: Vec<String>,
}

// TODO load in a hashMap
// #[allow(clippy::missing_docs_in_private_items)]
// #[allow(non_snake_case)]
// #[derive(Clone, Debug, Serialize,  Deserialize, Default)]
// pub struct PkgsConfig {
//   pub haproxy: Option<Haproxy>,
//   pub coredns: Option<Coredns>,
//   pub etcd: Option<Etcd>,
//   pub keepalived: Option<Keepalived>,
//   pub name: String,
//   pub info: PkgIn,
// }

#[allow(clippy::missing_docs_in_private_items)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct PkgInfo {
  pub vers: String,
  pub arch: Option<String>,
  pub file: Option<String>,
  pub url: Option<String>,
  pub url2: Option<String>,
  pub prd: Option<String>,
  pub pkg: Option<String>,
}

#[allow(clippy::missing_docs_in_private_items)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct PkgVers {
  pub vers: String,
  pub prd: Option<String>,
  pub install: Option<String>,
  pub data: Option<String>,
  pub path: Option<String>,
}
/*
#[allow(clippy::missing_docs_in_private_items)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct Kubernetes {
  pub k8s: PkgInfo,
  pub pkgs: Vec<K8sPkgs>,
}
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct K8sPkgs {
  pub name: String,
  pub info: PkgVers,
}
// TODO load in a hashMap
/*
  pub dashboard: PkgIn,
  pub cillium: PkgIn,
  pub helm: PkgIn,
  pub istio: PkgIn,
  pub rook: PkgIn,
  pub ceph: PkgIn,
  pub tekton: PkgIn,
  pub tektonDashboard: PkgIn,
  pub tikv: PkgIn,
  pub titan: PkgIn,
}
*/

#[allow(clippy::missing_docs_in_private_items)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct PkgVers {
  pub name: String,
  pub info: PkgIn,
}
*/
#[allow(clippy::missing_docs_in_private_items)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct Packages {
  pub pkgs: Vec<PkgVers>,
}
// TODO load in a hashMap
// #[allow(clippy::missing_docs_in_private_items)]
// #[derive(Clone, Debug, Serialize,  Deserialize, Default)]
// pub struct PkgsVers {
//   pub containerd: PkgInfo,
//   pub crictl: PkgInfo,
//   pub coredns: PkgInfo,
//   pub etcd: PkgInfo,
//   pub haproxy: PkgInfo,
//   pub keepalived: PkgInfo,
//   pub restic: PkgInfo,
//   pub yq: PkgInfo,
//   pub s3cmd: PkgInfo,
//   pub kubernetes: Kubernetes, 
// }

pub fn load_pack_info(tsksrvc_vars: &mut HashMap<String,String>, package: &str, pkg_info: PkgInfo ) {
  tsksrvc_vars.insert(format!("{}_vers",package),pkg_info.vers);
  tsksrvc_vars.insert(format!("{}_arch",package),pkg_info.arch.unwrap_or(String::from("")));
  tsksrvc_vars.insert(format!("{}_file",package),pkg_info.file.unwrap_or(String::from("")));
  tsksrvc_vars.insert(format!("{}_url",package),pkg_info.url.unwrap_or(String::from("")));
  tsksrvc_vars.insert(format!("{}_url2",package),pkg_info.url2.unwrap_or(String::from("")));
  tsksrvc_vars.insert(format!("{}_prd",package),pkg_info.prd.unwrap_or(String::from("")));
  tsksrvc_vars.insert(format!("{}_pkg",package),pkg_info.pkg.unwrap_or(String::from("")));
}

pub async fn get_pkgs_vers(cloud: &mut Cloud) -> Result<()> {
	let env_path: String;
	if Path::new(&cloud.env.versions).exists() {
		env_path=String::from(&cloud.env.versions);
	} else {
		env_path=format!("{}/{}",&cloud.env.path,&cloud.env.versions);
	}
  let vers_content= fs::read_to_string(&env_path).with_context(|| format!("Failed to read from {}", &cloud.env.versions))?;
  cloud.pkgs = serde_yaml::from_str(&vers_content)?;
  Ok(())
}

pub fn load_pack_vars(tsksrvc_vars: &mut HashMap<String,String>, cloud: &Cloud, pkg_path: &str) -> Result<()> {
  let pkg_data = fs::read_to_string(&pkg_path)
  .with_context(|| format!("Failed to read 'pkg_path' from {}", &pkg_path))?;  
  let pkg_vars: HashMap<String,String> = serde_yaml::from_str(&pkg_data)?;
  for (key,value) in pkg_vars.iter() {
    tsksrvc_vars.insert(key.to_owned(), value.to_owned());
    if key == "req_packs" {
      for req_pack in value.split(" ") {
        if let Some(pk) = cloud.pkgs.get(req_pack) {
          load_pack_info(tsksrvc_vars, &req_pack, pk.to_owned());
        }
      }
    }
  }
  Ok(())
}

pub fn load_pack_config(tsksrvc_vars: &mut HashMap<String,String>, cloud: &Cloud, tsk: &TskSrvc, hostname: &str) -> Result<()> {
  // print!(" {} ",&tsk.name);
  let tsk_name = format!("{}",&tsk.name);
  tsksrvc_vars.insert("pack_name".to_string(),tsk_name.to_owned());
  if let Some(pkg) = cloud.pkgs.get(&tsk_name) {
    load_pack_info(tsksrvc_vars, &tsk_name, pkg.to_owned());
  }
  // let mut pkg_path = format!("{}/home/{}/packs/{}_{}.yaml",&cloud.env.path,&cloud.env.source_pathh,&hostname,&tsk.name);
  let mut pkg_path = format!("{}/packs/{}_{}.yaml",&cloud.env.source_path,&hostname,&tsk.name);
  if Path::new(&pkg_path).exists() {
    load_pack_vars(tsksrvc_vars,cloud,&pkg_path)?;
  }
  pkg_path = format!("{}/packs/{}.yaml",&cloud.env.source_path,&tsk.name);
  if Path::new(&pkg_path).exists() {
    load_pack_vars(tsksrvc_vars,cloud,&pkg_path)?;
  }
  pkg_path = format!("{}/{}/{}.yaml",&cloud.env.tsksrvcs_path,&tsk.name,&tsk.name);
  if Path::new(&pkg_path).exists() {
    load_pack_vars(tsksrvc_vars,cloud,&pkg_path)?;
  }
  Ok(())
}